#!/usr/bin/env python
"""
Scummbler v2

    This file is part of Scummbler.

    Scummbler is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Scummbler is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Scummbler.  If not, see <http://www.gnu.org/licenses/>.

Laurence Dougal Myers
www.jestarjokin.net

December 2008 - May 2009

TODO: add support for v3 and v4 games, possibly FM-Towns

Differences in older engines:

v3, v4: actorOps - Scale(p8)
v3: cursorCommand - LoadCharset(p8, p8) is $0E
v3, v4: drawObject(p16, p16, p16), is $25 (overrides pickupObject)
indy3: waitForActor is $3B (overrides getActorScale)
indy3: getActorX and getActorY take p8
v3, v4: ifState is new, $0F (overrides getObjectState), ifNotState is new, $2F
v3: setBoxFlags(p8, b) is $30 (overrides matrixOps)
v3?, v4: oldRoomEffect-set and oldRoomEffect-fadein is new, $5C
v3, v4: pickupObject(p16) (old) is $50
v3: print and printEgo subop $06 is Height(p16)
LoomCD v4: print and printEgo adds subop $08 PlayCDTrack(p16, p16)
FM-Towns: ResourceRoutines adds some sub-ops
v3: roomOps stores arguments before sub-op
v3, v4: roomOps adds sub-op $02 ("RoomColor")
v3, v4: roomOps $04 SetPalColor only takes the two arguments
v3, v4: saveLoadGame is $22 (overrides getAnimCounter)
v3, v4: saveLoadVars is $A7 (overrides dummy)
small header: WaitForSentence is $4C (overrides soundKludge),
indy3: WaitForMessage is $AE (overrides wait with sub-ops)
"""
from pyparsing import *
from scummbler_lexicon import *
from scummbler_opcodes import *
from scummbler_misc import SupLit, NamedElement


class GrammarBase(object):
    # These two need parse actions assigned, so we re-add them as elements of this class,
    #  to join (and be assigned like) the rest of the grammar.
    leString = leString
    leDefinedValue = leDefinedValue
    
    # This is the expression that compiler will call to parse strings.
    # It's a string so that our iterator doesn't return it as part of the grammar.
    # It needs to be overridden by any inheriting grammar.
    rootExpression = 'ParserElement()'
    testExpression = 'ParserElement()' # used for testing, similar to above
    
    def __iter__(self):
        """ Iterating a grammar returns a tuple containing the name of the grammar element and the actual object."""
        return ((g, getattr(self, g)) for g in dir(self.__class__) if issubclass(getattr(self, g).__class__, ParserElement))
        

class GrammarV5(GrammarBase):
    # ~~~~~~~~~~~~ AUTO-GENERATED FUNCTIONS ~~~~~~~~~~~~~
    grFunc_actorFollowCamera = Literal("actorFollowCamera")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_actorFromPos = Group(leVariable)("target") + SupLit("=") + Literal("actorFromPos")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(",") + Group(leVarOrWord)("arg2") + SupLit(")")
    
    # --- Start sub-opcodes ---
    grFunc_ActorOps_Unknown = Literal("Unknown")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_ActorOps_Costume = Literal("Costume")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_ActorOps_WalkSpeed = Literal("WalkSpeed")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + Group(leVarOrByte)("arg2") + SupLit(")")
    grFunc_ActorOps_Sound = Literal("Sound")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_ActorOps_WalkAnimNr = Literal("WalkAnimNr")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_ActorOps_TalkAnimNr = Literal("TalkAnimNr")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + Group(leVarOrByte)("arg2") + SupLit(")")
    grFunc_ActorOps_StandAnimNr = Literal("StandAnimNr")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_ActorOps_Nothing = Literal("Nothing")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + Group(leVarOrByte)("arg2") + SupLit(",") + Group(leVarOrByte)("arg3") + SupLit(")")
    grFunc_ActorOps_Init = Literal("Init")("function") + SupLit("(") + SupLit(")")
    grFunc_ActorOps_Elevation = Literal("Elevation")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(")")
    grFunc_ActorOps_DefaultAnims = Literal("DefaultAnims")("function") + SupLit("(") + SupLit(")")
    grFunc_ActorOps_Palette = Literal("Palette")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + Group(leVarOrByte)("arg2") + SupLit(")")
    grFunc_ActorOps_TalkColor = Literal("TalkColor")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_ActorOps_Name = Literal("Name")("function") + SupLit("(") + NamedElement(leString, "arg1") + SupLit(")")
    grFunc_ActorOps_InitAnimNr = Literal("InitAnimNr")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_ActorOps_Width = Literal("Width")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_ActorOps_Scale = Literal("Scale")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + Group(leVarOrByte)("arg2") + SupLit(")")
    grFunc_ActorOps_NeverZClip = Literal("NeverZClip")("function") + SupLit("(") + SupLit(")")
    grFunc_ActorOps_SetZClip = Literal("SetZClip")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_ActorOps_IgnoreBoxes = Literal("IgnoreBoxes")("function") + SupLit("(") + SupLit(")")
    grFunc_ActorOps_FollowBoxes = Literal("FollowBoxes")("function") + SupLit("(") + SupLit(")")
    grFunc_ActorOps_AnimSpeed = Literal("AnimSpeed")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_ActorOps_ShadowMode = Literal("ShadowMode")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_ActorOps_SO = grFunc_ActorOps_Unknown | grFunc_ActorOps_Costume | grFunc_ActorOps_WalkSpeed | grFunc_ActorOps_Sound | grFunc_ActorOps_WalkAnimNr | grFunc_ActorOps_TalkAnimNr | grFunc_ActorOps_StandAnimNr | grFunc_ActorOps_Nothing | grFunc_ActorOps_Init | grFunc_ActorOps_Elevation | grFunc_ActorOps_DefaultAnims | grFunc_ActorOps_Palette | grFunc_ActorOps_TalkColor | grFunc_ActorOps_Name | grFunc_ActorOps_InitAnimNr | grFunc_ActorOps_Width | grFunc_ActorOps_Scale | grFunc_ActorOps_NeverZClip | grFunc_ActorOps_SetZClip | grFunc_ActorOps_IgnoreBoxes | grFunc_ActorOps_FollowBoxes | grFunc_ActorOps_AnimSpeed | grFunc_ActorOps_ShadowMode
    # --- End sub-opcodes ---
    grFunc_ActorOps = Literal("ActorOps")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + SupLit("[") + Optional(delimitedList(grFunc_ActorOps_SO))("arg2") + SupLit("]")  + SupLit(")")

    grFunc_setClass = Literal("setClass")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(",") + leArgList("arg2") + SupLit(")")
    grFunc_animateCostume = Literal("animateCostume")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + Group(leVarOrByte)("arg2") + SupLit(")")
    grFunc_breakHere = Literal("breakHere")("function") + SupLit("(") + SupLit(")")
    grFunc_chainScript = Literal("chainScript")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + leArgList("arg2") + SupLit(")")
    
    # --- Start sub-opcodes ---
    grFunc_cursorCommand_CursorShow = Literal("CursorShow")("function") + SupLit("(") + SupLit(")")
    grFunc_cursorCommand_CursorHide = Literal("CursorHide")("function") + SupLit("(") + SupLit(")")
    grFunc_cursorCommand_UserputOn = Literal("UserputOn")("function") + SupLit("(") + SupLit(")")
    grFunc_cursorCommand_UserputOff = Literal("UserputOff")("function") + SupLit("(") + SupLit(")")
    grFunc_cursorCommand_CursorSoftOn = Literal("CursorSoftOn")("function") + SupLit("(") + SupLit(")")
    grFunc_cursorCommand_CursorSoftOff = Literal("CursorSoftOff")("function") + SupLit("(") + SupLit(")")
    grFunc_cursorCommand_UserputSoftOn = Literal("UserputSoftOn")("function") + SupLit("(") + SupLit(")")
    grFunc_cursorCommand_UserputSoftOff = Literal("UserputSoftOff")("function") + SupLit("(") + SupLit(")")
    grFunc_cursorCommand_SetCursorImg = Literal("SetCursorImg")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + Group(leVarOrByte)("arg2") + SupLit(")")
    grFunc_cursorCommand_setCursorHotspot = Literal("setCursorHotspot")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + Group(leVarOrByte)("arg2") + SupLit(",") + Group(leVarOrByte)("arg3") + SupLit(")")
    grFunc_cursorCommand_InitCursor = Literal("InitCursor")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_cursorCommand_InitCharset = Literal("InitCharset")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_cursorCommand_CursorCommand = Literal("CursorCommand")("function") + SupLit("(") + leArgList("arg1") + SupLit(")")
    grFunc_cursorCommand_SO = grFunc_cursorCommand_CursorShow | grFunc_cursorCommand_CursorHide | grFunc_cursorCommand_UserputOn | grFunc_cursorCommand_UserputOff | grFunc_cursorCommand_CursorSoftOn | grFunc_cursorCommand_CursorSoftOff | grFunc_cursorCommand_UserputSoftOn | grFunc_cursorCommand_UserputSoftOff | grFunc_cursorCommand_SetCursorImg | grFunc_cursorCommand_setCursorHotspot | grFunc_cursorCommand_InitCursor | grFunc_cursorCommand_InitCharset | grFunc_cursorCommand_CursorCommand
    # --- End sub-opcodes ---
    
    grFunc_cutscene = Literal("cutscene")("function") + SupLit("(") + leArgList("arg1") + SupLit(")")
    grFunc_debug = Literal("debug?")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(")")
    grFunc_delayVariable = Literal("delayVariable")("function") + SupLit("(") + Group(leVariable)("arg1") + SupLit(")")
    grFunc_dummyA7 = Literal("dummy(A7)")("function") + SupLit("(") + SupLit(")")
    grFunc_endCutscene = Literal("endCutscene")("function") + SupLit("(") + SupLit(")")
    grFunc_faceActor = Literal("faceActor")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + Group(leVarOrWord)("arg2") + SupLit(")")
    grFunc_findInventory = Group(leVariable)("target") + SupLit("=") + Literal("findInventory")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + Group(leVarOrByte)("arg2") + SupLit(")")
    grFunc_findObject = Group(leVariable)("target") + SupLit("=") + Literal("findObject")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + Group(leVarOrByte)("arg2") + SupLit(")")
    grFunc_freezeScripts = Literal("freezeScripts")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_getActorCostume = Group(leVariable)("target") + SupLit("=") + Literal("getActorCostume")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_getActorElevation = Group(leVariable)("target") + SupLit("=") + Literal("getActorElevation")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_getActorFacing = Group(leVariable)("target") + SupLit("=") + Literal("getActorFacing")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_getActorMoving = Group(leVariable)("target") + SupLit("=") + Literal("getActorMoving")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_getActorRoom = Group(leVariable)("target") + SupLit("=") + Literal("getActorRoom")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_getActorScale = Group(leVariable)("target") + SupLit("=") + Literal("getActorScale")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_getActorWalkBox = Group(leVariable)("target") + SupLit("=") + Literal("getActorWalkBox")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_getActorWidth = Group(leVariable)("target") + SupLit("=") + Literal("getActorWidth")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_getActorX = Group(leVariable)("target") + SupLit("=") + Literal("getActorX")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(")")
    grFunc_getActorY = Group(leVariable)("target") + SupLit("=") + Literal("getActorY")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(")")
    grFunc_getAnimCounter = Group(leVariable)("target") + SupLit("=") + Literal("getAnimCounter")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_getClosestObjActor = Group(leVariable)("target") + SupLit("=") + Literal("getClosestObjActor")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(")")
    grFunc_getDist = Group(leVariable)("target") + SupLit("=") + Literal("getDist")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(",") + Group(leVarOrWord)("arg2") + SupLit(")")
    grFunc_getInventoryCount = Group(leVariable)("target") + SupLit("=") + Literal("getInventoryCount")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_getObjectOwner = Group(leVariable)("target") + SupLit("=") + Literal("getObjectOwner")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(")")
    grFunc_getObjectState = Group(leVariable)("target") + SupLit("=") + Literal("getObjectState")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(")")
    grFunc_getRandomNr = Group(leVariable)("target") + SupLit("=") + Literal("getRandomNr")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_getStringWidth = Group(leVariable)("target") + SupLit("=") + Literal("getStringWidth")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_getVerbEntryPoint = Group(leVariable)("target") + SupLit("=") + Literal("getVerbEntryPoint")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(",") + Group(leVarOrWord)("arg2") + SupLit(")")
    grFunc_isScriptRunning = Group(leVariable)("target") + SupLit("=") + Literal("isScriptRunning")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_isSoundRunning = Group(leVariable)("target") + SupLit("=") + Literal("isSoundRunning")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_lights = Literal("lights")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + Group(leByte)("arg2") + SupLit(",") + Group(leByte)("arg3") + SupLit(")")
    grFunc_loadRoom = Literal("loadRoom")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    
    # --- Start sub-opcodes ---
    grFunc_matrixOp_setBoxFlags = Literal("setBoxFlags")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + Group(leVarOrByte)("arg2") + SupLit(")")
    grFunc_matrixOp_setBoxScale = Literal("setBoxScale")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + Group(leVarOrByte)("arg2") + SupLit(")")
    grFunc_matrixOp_SetBoxSlot = Literal("SetBoxSlot")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + Group(leVarOrByte)("arg2") + SupLit(")")
    grFunc_matrixOp_createBoxMatrix = Literal("createBoxMatrix")("function") + SupLit("(") + SupLit(")")
    grFunc_matrixOp_SO = grFunc_matrixOp_setBoxFlags | grFunc_matrixOp_setBoxScale | grFunc_matrixOp_SetBoxSlot | grFunc_matrixOp_createBoxMatrix
    # --- End sub-opcodes ---
    
    grFunc_panCameraTo = Literal("panCameraTo")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(")")
    
    # --- Start sub-opcodes ---
    grFunc_print_Pos = Literal("Pos")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(",") + Group(leVarOrWord)("arg2") + SupLit(")")
    grFunc_print_Color = Literal("Color")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_print_Clipped = Literal("Clipped")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(")")
    grFunc_print_RestoreBG = Literal("RestoreBG")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(",") + Group(leVarOrWord)("arg2") + SupLit(")")
    grFunc_print_Center = Literal("Center")("function") + SupLit("(") + SupLit(")")
    grFunc_print_Left = Literal("Left")("function") + SupLit("(") + SupLit(")")
    grFunc_print_Overhead = Literal("Overhead")("function") + SupLit("(") + SupLit(")")
    grFunc_print_PlayCDTrack = Literal("PlayCDTrack")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(",") + Group(leVarOrWord)("arg2") + SupLit(")")
    grFunc_print_Text = Literal("Text")("function") + SupLit("(") + NamedElement(leString, "arg1") + SupLit(")")
    grFunc_print_SO = grFunc_print_Pos | grFunc_print_Color | grFunc_print_Clipped | grFunc_print_RestoreBG | grFunc_print_Center | grFunc_print_Left | grFunc_print_Overhead | grFunc_print_PlayCDTrack | grFunc_print_Text
    # --- End sub-opcodes ---
    grFunc_print = Literal("print")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + SupLit("[") + Optional(delimitedList(grFunc_print_SO))("arg2") + SupLit("]")  + SupLit(")")
    grFunc_putActor = Literal("putActor")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + Group(leVarOrWord)("arg2") + SupLit(",") + Group(leVarOrWord)("arg3") + SupLit(")")
    grFunc_putActorAtObject = Literal("putActorAtObject")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + Group(leVarOrWord)("arg2") + SupLit(")")
    grFunc_putActorInRoom = Literal("putActorInRoom")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + Group(leVarOrByte)("arg2") + SupLit(")")
    
    # --- Start sub-opcodes ---
    grFunc_Resource_ResourceloadScript = Literal("Resource.loadScript")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_Resource_ResourceloadSound = Literal("Resource.loadSound")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_Resource_ResourceloadCostume = Literal("Resource.loadCostume")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_Resource_ResourceloadRoom = Literal("Resource.loadRoom")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_Resource_ResourcenukeScript = Literal("Resource.nukeScript")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_Resource_ResourcenukeSound = Literal("Resource.nukeSound")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_Resource_ResourcenukeCostume = Literal("Resource.nukeCostume")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_Resource_ResourcenukeRoom = Literal("Resource.nukeRoom")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_Resource_ResourcelockScript = Literal("Resource.lockScript")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_Resource_ResourcelockSound = Literal("Resource.lockSound")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_Resource_ResourcelockCostume = Literal("Resource.lockCostume")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_Resource_ResourcelockRoom = Literal("Resource.lockRoom")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_Resource_ResourceunlockScript = Literal("Resource.unlockScript")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_Resource_ResourceunlockSound = Literal("Resource.unlockSound")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_Resource_ResourceunlockCostume = Literal("Resource.unlockCostume")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_Resource_ResourceunlockRoom = Literal("Resource.unlockRoom")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_Resource_ResourceclearHeap = Literal("Resource.clearHeap")("function") + SupLit("(") + SupLit(")")
    grFunc_Resource_ResourceloadCharset = Literal("Resource.loadCharset")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_Resource_ResourcenukeCharset = Literal("Resource.nukeCharset")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_Resource_ResourceloadFlObject = Literal("Resource.loadFlObject")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + Group(leVarOrWord)("arg2") + SupLit(")")
    grFunc_Resource_SO = grFunc_Resource_ResourceloadScript | grFunc_Resource_ResourceloadSound | grFunc_Resource_ResourceloadCostume | grFunc_Resource_ResourceloadRoom | grFunc_Resource_ResourcenukeScript | grFunc_Resource_ResourcenukeSound | grFunc_Resource_ResourcenukeCostume | grFunc_Resource_ResourcenukeRoom | grFunc_Resource_ResourcelockScript | grFunc_Resource_ResourcelockSound | grFunc_Resource_ResourcelockCostume | grFunc_Resource_ResourcelockRoom | grFunc_Resource_ResourceunlockScript | grFunc_Resource_ResourceunlockSound | grFunc_Resource_ResourceunlockCostume | grFunc_Resource_ResourceunlockRoom | grFunc_Resource_ResourceclearHeap | grFunc_Resource_ResourceloadCharset | grFunc_Resource_ResourcenukeCharset | grFunc_Resource_ResourceloadFlObject
    # --- End sub-opcodes ---
    
    grFunc_setCameraAt = Literal("setCameraAt")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(")")
    grFunc_setObjectName = Literal("setObjectName")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(",") + NamedElement(leString, "arg2") + SupLit(")")
    grFunc_setOwnerOf = Literal("setOwnerOf")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(",") + Group(leVarOrByte)("arg2") + SupLit(")")
    grFunc_setState = Literal("setState")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(",") + Group(leVarOrByte)("arg2") + SupLit(")")
    grFunc_soundKludge = Literal("soundKludge")("function") + SupLit("(") + leArgList("arg1") + SupLit(")")
    grFunc_startMusic = Literal("startMusic")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_startObject = Literal("startObject")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(",") + Group(leVarOrByte)("arg2") + SupLit(",") + leArgList("arg3") + SupLit(")")
    grFunc_startSound = Literal("startSound")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_stopMusic = Literal("stopMusic")("function") + SupLit("(") + SupLit(")")
    grFunc_stopObjectScript = Literal("stopObjectScript")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(")")
    grFunc_stopScript = Literal("stopScript")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_stopSound = Literal("stopSound")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    
    # --- Start sub-opcodes ---
    grFunc_stringOps_PutCodeInString = Literal("PutCodeInString")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + NamedElement(leString, "arg2") + SupLit(")")
    grFunc_stringOps_CopyString = Literal("CopyString")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + Group(leVarOrByte)("arg2") + SupLit(")")
    grFunc_stringOps_SetStringChar = Literal("SetStringChar")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + Group(leVarOrByte)("arg2") + SupLit(",") + Group(leVarOrByte)("arg3") + SupLit(")")
    grFunc_stringOps_GetStringChar = Group(leVariable)("target") + SupLit("=") + Literal("GetStringChar")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + Group(leVarOrByte)("arg2") + SupLit(")")
    grFunc_stringOps_CreateString = Literal("CreateString")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + Group(leVarOrByte)("arg2") + SupLit(")")
    grFunc_stringOps_SO = grFunc_stringOps_PutCodeInString | grFunc_stringOps_CopyString | grFunc_stringOps_SetStringChar | grFunc_stringOps_GetStringChar | grFunc_stringOps_CreateString
    # --- End sub-opcodes ---

    # --- Start sub-opcodes ---
    grFunc_VerbOps_Image = Literal("Image")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(")")
    grFunc_VerbOps_Text = Literal("Text")("function") + SupLit("(") + NamedElement(leString, "arg1") + SupLit(")")
    grFunc_VerbOps_Color = Literal("Color")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_VerbOps_HiColor = Literal("HiColor")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_VerbOps_SetXY = Literal("SetXY")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(",") + Group(leVarOrWord)("arg2") + SupLit(")")
    grFunc_VerbOps_On = Literal("On")("function") + SupLit("(") + SupLit(")")
    grFunc_VerbOps_Off = Literal("Off")("function") + SupLit("(") + SupLit(")")
    grFunc_VerbOps_Delete = Literal("Delete")("function") + SupLit("(") + SupLit(")")
    grFunc_VerbOps_New = Literal("New")("function") + SupLit("(") + SupLit(")")
    grFunc_VerbOps_DimColor = Literal("DimColor")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_VerbOps_Dim = Literal("Dim")("function") + SupLit("(") + SupLit(")")
    grFunc_VerbOps_Key = Literal("Key")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_VerbOps_Center = Literal("Center")("function") + SupLit("(") + SupLit(")")
    grFunc_VerbOps_SetToString = Literal("SetToString")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(")")
    grFunc_VerbOps_SetToObject = Literal("SetToObject")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(",") + Group(leVarOrByte)("arg2") + SupLit(")")
    grFunc_VerbOps_BackColor = Literal("BackColor")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_VerbOps_SO = grFunc_VerbOps_Image | grFunc_VerbOps_Text | grFunc_VerbOps_Color | grFunc_VerbOps_HiColor | grFunc_VerbOps_SetXY | grFunc_VerbOps_On | grFunc_VerbOps_Off | grFunc_VerbOps_Delete | grFunc_VerbOps_New | grFunc_VerbOps_DimColor | grFunc_VerbOps_Dim | grFunc_VerbOps_Key | grFunc_VerbOps_Center | grFunc_VerbOps_SetToString | grFunc_VerbOps_SetToObject | grFunc_VerbOps_BackColor
    # --- End sub-opcodes ---
    grFunc_VerbOps = Literal("VerbOps")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + SupLit("[") + Optional(delimitedList(grFunc_VerbOps_SO))("arg2") + SupLit("]")  + SupLit(")")
    
    # --- Start sub-opcodes ---
    grFunc_wait_WaitForActor = Literal("WaitForActor")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(")")
    grFunc_wait_WaitForMessage = Literal("WaitForMessage")("function") + SupLit("(") + SupLit(")")
    grFunc_wait_WaitForCamera = Literal("WaitForCamera")("function") + SupLit("(") + SupLit(")")
    grFunc_wait_WaitForSentence = Literal("WaitForSentence")("function") + SupLit("(") + SupLit(")")
    grFunc_wait_SO = grFunc_wait_WaitForActor | grFunc_wait_WaitForMessage | grFunc_wait_WaitForCamera | grFunc_wait_WaitForSentence
    # --- End sub-opcodes ---
    
    
    grFunc_walkActorTo = Literal("walkActorTo")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + Group(leVarOrWord)("arg2") + SupLit(",") + Group(leVarOrWord)("arg3") + SupLit(")")
    grFunc_walkActorToActor = Literal("walkActorToActor")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + Group(leVarOrByte)("arg2") + SupLit(",") + Group(leByte)("arg3") + SupLit(")")
    grFunc_walkActorToObject = Literal("walkActorToObject")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + Group(leVarOrWord)("arg2") + SupLit(")")
    grFunctions = grFunc_actorFollowCamera | grFunc_actorFromPos | grFunc_ActorOps | grFunc_setClass | grFunc_animateCostume | grFunc_breakHere | grFunc_chainScript | grFunc_cursorCommand_SO | grFunc_cutscene | grFunc_debug | grFunc_delayVariable | grFunc_dummyA7 | grFunc_endCutscene | grFunc_faceActor | grFunc_findInventory | grFunc_findObject | grFunc_freezeScripts | grFunc_getActorCostume | grFunc_getActorElevation | grFunc_getActorFacing | grFunc_getActorMoving | grFunc_getActorRoom | grFunc_getActorScale | grFunc_getActorWalkBox | grFunc_getActorWidth | grFunc_getActorX | grFunc_getActorY | grFunc_getAnimCounter | grFunc_getClosestObjActor | grFunc_getDist | grFunc_getInventoryCount | grFunc_getObjectOwner | grFunc_getObjectState | grFunc_getRandomNr | grFunc_getStringWidth | grFunc_getVerbEntryPoint | grFunc_isScriptRunning | grFunc_isSoundRunning | grFunc_lights | grFunc_loadRoom | grFunc_matrixOp_SO | grFunc_panCameraTo | grFunc_print | grFunc_putActor | grFunc_putActorAtObject | grFunc_putActorInRoom | grFunc_Resource_SO | grFunc_setCameraAt | grFunc_setObjectName | grFunc_setOwnerOf | grFunc_setState | grFunc_soundKludge | grFunc_startMusic | grFunc_startObject | grFunc_startSound | grFunc_stopMusic | grFunc_stopObjectScript | grFunc_stopScript | grFunc_stopSound | grFunc_stringOps_SO | grFunc_VerbOps | grFunc_wait_SO | grFunc_walkActorTo | grFunc_walkActorToActor | grFunc_walkActorToObject
    # ~~~~~~~~~~~~ END AUTO-GENERATED FUNCTIONS ~~~~~~~~~~~~~
    
    
    
    # Inline opcodes (not function calls)
    grInlineOperation = Group(leVariable)("target") + oneOf("+= -= /= *= &= |= =")("operation") + Group(leVarOrWord)("value")
    grIncDec = Group(leVariable)("target") + oneOf("++ --")("operation") # not sure if Bits can be incremented/decremented
    
    # Jumps
    grJumpLabel = SupLit("[") + leLabel("label") + SupLit("]")
    grJumpGoto = SupLit("goto") + leLabel("target")
    grJumpExpression = grJumpGoto
    
    # Comparison & conditional jump/"if" related opcodes
    grIfComparison = Group(leVariable)("arg1") + oneOf(" ".join(opComparisons.keys()))("comparator") + Group(leVarOrWord)("arg2")
    grIfZeroEquality = Optional(Literal("!")("!")) + leVariable
    grIfClassOfIs = Literal("classOfIs")("function") + SupLit("(") + Group(leVarOrWord)("testobject") + SupLit(",") + leArgList("classes") ^ Optional(SupLit(")")) # descumm forgets to add the last parenthesis
    grIfActorInBox = Literal("isActorInBox")("function") + SupLit("(") + Group(leVarOrByte)("actor") + SupLit(",") + Group(leVarOrByte)("box") + SupLit(")")
    grIfState = Literal("getState")("function") + SupLit("(") + Group(leVarOrWord)("testobject") + SupLit(")") + (Literal("==") | Literal("!="))("comparator") + Group(leVarOrByte)("state") + SupLit(")")
    grBooleanExpression = grIfComparison ^ grIfZeroEquality ^ grIfClassOfIs ^ grIfActorInBox ^ grIfState
    grJumpUnless = SupLit("unless") + SupLit("(") + (grBooleanExpression)("test") + SupLit(")") + SupLit("goto") + leLabel("target")
    grIfStart = SupLit("if") + SupLit("(") + (grBooleanExpression) + SupLit(")") + SupLit("{")
    grIfElse = SupLit("}") + SupLit("else")
    grIfEnd = SupLit("}")
    
    # Other control flow
    grWhileStart = SupLit("while") + SupLit("(") + grBooleanExpression + SupLit(")") + SupLit("{")
    grWhileEnd = SupLit("}") # obviously this conflicts with grIfEnd...
    grDoWhileStart = SupLit("do") + SupLit("{")
    grDoWhileEnd = SupLit("}") + SupLit("while") + SupLit("(") + grBooleanExpression + SupLit(")")
    grForLoopStart = (SupLit("for") + SupLit("(") +
                      Optional(grInlineOperation)("init") + SupLit(";") + 
                      Group(grBooleanExpression)("test") + SupLit(";") + 
                      Optional(grInlineOperation | grIncDec)("increm") +
                      SupLit(")") + SupLit("{")) # "For" loops only allow basic inline maths for now
    grForLoopEnd = SupLit("}")
    
    
    # This used to be below but I need it for startScript
    grDescummOpcode = SupLit("(") + (leHexByte | leLabel | Literal("**"))("dop") + SupLit(")")
    
    # Functions that can't be automatically generated due to quirks in their grammar.
    
    # delay has its own type of argument, a 24-bit constant
    grFunc_delay = Literal("delay")("function") + SupLit("(") + Group(leDelay)("arg1") + SupLit(")")
    # doSentence normally takes 3 params, except when it only takes 1 (which is 0xFE, descumm outputs it as "STOP")
    grFunc_doSentence = (Literal("doSentence")("function") + SupLit("(") +
                         (Literal("STOP")("arg1") | (Group(leVarOrByte)("arg1") + SupLit(",") +
                                                     Group(leVarOrWord)("arg2") + SupLit(",") +
                                                     Group(leVarOrWord)("arg3")
                                                     )
                         ) + SupLit(")"))
    # drawBox has aux opcodes
    grFunc_drawBox = ( Literal("drawBox")("function") + SupLit("(") +
                      Group(leVarOrWord)("left") + SupLit(",") + 
                      Group(leVarOrWord)("top") + SupLit(",") + 
                      Group(leVarOrWord)("right") + SupLit(",") +
                      Group(leVarOrWord)("bottom") + SupLit(",") +
                      Group(leVarOrByte)("colour") + SupLit(")") )
    # drawObject omits opening bracket of 0x1F (unnamed sub-opcode) but accidentally leaves in the closing bracket.
    # Unfortunately, I have to define the whole of drawObject here (although it's based on auto-generated code).
    grFunc_drawObject_setXY = Literal("setXY")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(",") + Group(leVarOrWord)("arg2") + SupLit(")")
    grFunc_drawObject_setImage = Literal("setImage")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(")")
    grFunc_drawObject_SO = grFunc_drawObject_setXY | grFunc_drawObject_setImage
    # Make the whole of "arg2" to support the strange "empty" function - if it's missing, parse action knows what to do
    grFunc_drawObject = (Literal("drawObject")("function") + SupLit("(") +
                         Group(leVarOrWord)("arg1") +
                         Optional( Suppress(Optional(Literal(","))) + grFunc_drawObject_SO("arg2") ) +
                         SupLit(")") + Optional(SupLit(")")))
    # loadRoomWithEgo has two params and two constants
    grFunc_loadRoomWithEgo = (Literal("loadRoomWithEgo")("function") + SupLit("(") +
                              Group(leVarOrWord)("object") + SupLit(",") +
                              Group(leVarOrByte)("room") + SupLit(",") +
                              leWord("x") + SupLit(",") +
                              leWord("y") +
                              SupLit(")"))
    # oldRoomEffect is dodgy
    grFunc_oldRoomEffect = ((Literal("oldRoomEffect-set") | Literal("oldRoomEffect-fadein"))("function") + SupLit("(") +
                            Group(leVarOrWord)("effect") +
                            SupLit(")")
                            )
    # override is like a soft bunny without ears (no parentheses) (except those) (and those)
    grFunc_override = (Literal("beginOverride") | Literal("endOverride"))("function") + Optional(SupLit("(") + SupLit(")"))
    # pickupObject can be one of two opcodes, one which takes a word and a byte, the other ("pickupObjectOld") only takes a word.
    grFunc_pickupObject = Literal("pickupObject")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + Optional( SupLit(",") + Group(leVarOrByte)("arg2") ) + SupLit(")")
    # "printEgo" re-uses the sub-opcodes for "print"
    grFunc_printEgo = Literal("printEgo")("function") + SupLit("(") + SupLit("[") + Optional(delimitedList(grFunc_print_SO))("arg1") + SupLit("]")  + SupLit(")")
    # pseudoRoom takes one constant followed by a variable number of constants.
    # Each res item is ORed with 0x80. A value of 0x80 is output as "IG" in descumm.
    grFunc_PseudoRoom = (Literal("PseudoRoom")("function") + SupLit("(") +
                         leByte("val") + SupLit(",") + 
                         delimitedList(leByte | Literal("IG"))("reslist") +
                         SupLit(")")
                        )
    # roomOps has sub-opcodes with non-standard encoding (mostly because of red, green, blue values)
    # These are auto-generated... (I have modified some to make the question marks optional)
    # A few sub-ops (RoomIntensity, SetScren) output an extra closing parenthesis
    grFunc_roomOps_RoomScroll = Literal("RoomScroll")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(",") + Group(leVarOrWord)("arg2") + SupLit(")")
    grFunc_roomOps_SetScreen = Literal("SetScreen")("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(",") + Group(leVarOrWord)("arg2") + SupLit(")")
    grFunc_roomOps_ShakeOn = Literal("ShakeOn")("function") + SupLit("(") + SupLit(")")
    grFunc_roomOps_ShakeOff = Literal("ShakeOff")("function") + SupLit("(") + SupLit(")")
    grFunc_roomOps_RoomIntensity = Literal("RoomIntensity")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + Group(leVarOrByte)("arg2") + SupLit(",") + Group(leVarOrByte)("arg3") + SupLit(")")
    grFunc_roomOps_saveLoad = (Literal("saveLoad?") ^ Literal("saveLoad"))("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + Group(leVarOrByte)("arg2") + SupLit(")")
    # screenEffect has a question mark and extra closing parenthesis
    grFunc_roomOps_screenEffect = (Literal("screenEffect?") ^ Literal("screenEffect"))("function") + SupLit("(") + Group(leVarOrWord)("arg1") + SupLit(")")
    grFunc_roomOps_colorCycleDelay = Literal("colorCycleDelay")("function") + SupLit("(") + Group(leVarOrByte)("arg1") + SupLit(",") + Group(leVarOrByte)("arg2") + SupLit(")")
    # Uses aux opcode, also extra closing parenthesis
    grFunc_roomOps_SetPalColor = (Literal("SetPalColor")("function") + SupLit("(") + 
                                  Group(leVarOrWord)("red") + SupLit(",") + 
                                  Group(leVarOrWord)("green") + SupLit(",") + 
                                  Group(leVarOrWord)("blue") + SupLit(",") + 
                                  Group(leVarOrByte)("index") +
                                  SupLit(")")
                                  )
    # Lots of parameters, descumm calls it "Unused" which (while possibly true) isn't very descriptive
    grFunc_roomOps_SetRoomScale = ((Literal("SetRoomScale") | Literal("Unused"))("function") + SupLit("(") + 
                             Group(leVarOrByte)("scale1") + SupLit(",") + 
                             Group(leVarOrByte)("y1") + SupLit(",") + 
                             Group(leVarOrByte)("scale2") + SupLit(",") + 
                             Group(leVarOrByte)("y2") + SupLit(",") + 
                             Group(leVarOrByte)("slot") + SupLit(",") + 
                             SupLit(")")
                             )
    # This has params and an aux opcode, but I'm also cheating and using this for setRoomShadow.
    grFunc_roomOps_setRGBRoomIntensity = ((Literal("setRGBRoomIntensity") | Literal("setRoomShadow"))("function") + SupLit("(") + 
                                          Group(leVarOrWord)("redscale") + SupLit(",") + 
                                          Group(leVarOrWord)("greenscale") + SupLit(",") + 
                                          Group(leVarOrWord)("bluescale") + SupLit(",") + 
                                          Group(leVarOrByte)("start") + SupLit(",") +
                                          Group(leVarOrByte)("end") +
                                          SupLit(")")
                                          )
    # Share grammar & parse action for saveString and loadString
    grFunc_roomOps_saveLoadString = ((Literal("saveString") | Literal("loadString"))("function") + SupLit("(") + 
                                     Group(leVarOrByte)("slot") + SupLit(",") + 
                                     NamedElement(leString, "stringy") +
                                     SupLit(")")
                                     )
    # Lots of params & aux opcodes
    grFunc_roomOps_palManipulate = (Literal("palManipulate")("function") + SupLit("(") + 
                                    Group(leVarOrByte)("slot") + SupLit(",") + 
                                    Group(leVarOrByte)("start") + SupLit(",") + 
                                    Group(leVarOrByte)("end") + SupLit(",") + 
                                    Group(leVarOrByte)("time") +
                                    SupLit(")"))
    # This big thing marks the end of roomOps, phew! Also, looks like a lot of roomOps add
    #  an extra closing parenthesis, so account for that here instead of in each sub-op.
    grFunc_roomOps_SO = (grFunc_roomOps_RoomScroll | grFunc_roomOps_SetScreen | grFunc_roomOps_ShakeOn | grFunc_roomOps_ShakeOff | 
                         grFunc_roomOps_RoomIntensity | grFunc_roomOps_saveLoad | grFunc_roomOps_screenEffect | 
                         grFunc_roomOps_colorCycleDelay | grFunc_roomOps_SetPalColor | grFunc_roomOps_SetRoomScale |
                         grFunc_roomOps_setRGBRoomIntensity | grFunc_roomOps_saveLoadString | grFunc_roomOps_palManipulate) + Optional(SupLit(")"))
    # setVarRange takes a starting variable, the number of constants that will be next in the instruction (I'm making this optional),
    #  and a delmitedList of either bytes or words (but not mixed). We determine which in the parse action.
    grFunc_setVarRange = (Literal("setVarRange")("function") + SupLit("(") +
                          Group(leVariable)("startvar") + SupLit(",") +
                          Optional(leByte("numitems") + SupLit(",")) +
                          SupLit("[") + delimitedList(leByte | leWord)("listvals") + SupLit("]") +
                          SupLit(")"))
    # saveRestoreVerbs - I originally misunderstood this instruction; I thought it had an aux opcode, three params, then a sub-opcode,
    #  when it actually just has a sub-opcode and three params. Still, this is more concise than the auto-generated instruction would be.
    grFunc_saveRestoreVerbs_saveVerbs = Literal("saveVerbs")
    grFunc_saveRestoreVerbs_restoreVerbs = Literal("restoreVerbs")
    grFunc_saveRestoreVerbs_deleteVerbs = Literal("deleteVerbs")
    grFunc_saveRestoreVerbs = ((grFunc_saveRestoreVerbs_saveVerbs | grFunc_saveRestoreVerbs_restoreVerbs | grFunc_saveRestoreVerbs_deleteVerbs)("function") + SupLit("(") +
                                Group(leVarOrByte)("start") + SupLit(",") +
                                Group(leVarOrByte)("end") + SupLit(",") +
                                Group(leVarOrByte)("mode") +
                                SupLit(")")
                                )
    # startScript modifies the opcode, but descumm doesn't 
    grFunc_startScript = (Literal("startScript")("function") + SupLit("(") +
                          Group(leVarOrByte)("arg1") + SupLit(",") +
                          leArgList("arg2") +
                          Optional(SupLit(",") + Literal("F"))("freezeres") + 
                          Optional(SupLit(",") + Literal("R"))("recursive") + 
                          SupLit(")"))
    # stopObjectCode must be the last item in a script
    grFunc_stopObjectCode = Literal("stopObjectCode")("function") + SupLit("(") + SupLit(")")
    # systemOps takes sub-opcodes, but they're only represented as the numbers 1, 2, or 3. I also allow meaningful words.
    grFunc_systemOps = (Literal("systemOps")("function") + SupLit("(") +
                        ((Literal("1") | Literal("RESTART")) | (Literal("2") | Literal("PAUSE")) | (Literal("3") | Literal("QUIT")))("subop") + # this is a bit crap
                        SupLit(")"))
    # ... here endeth the quirky functions
    
    # Pragramas (compiler directives)
    grPragmaScriptNum = Literal("Script#") + leUnsignedInt("scriptnum")
    grPragmaDefine = SupLit("#define") + leDefinedValueDef + Optional(SupLit("=")) + Group(leVarOrWord)
    grPragma = grPragmaScriptNum | grPragmaDefine
    
    # All quirky functions
    grQuirkyFunctions = (grFunc_printEgo | grFunc_delay | grFunc_doSentence | grFunc_drawBox | grFunc_drawObject | grFunc_loadRoomWithEgo | grFunc_pickupObject |
                         grFunc_PseudoRoom | grFunc_oldRoomEffect | grFunc_override | grFunc_setVarRange | grFunc_saveRestoreVerbs | grFunc_systemOps | 
                         grFunc_roomOps_SO)
    
    # Expression mode stuff
    grSubExpression = Forward()
    grExpressionOpcode = SupLit("<") + (grQuirkyFunctions | grFunctions)("expop") + SupLit(">") # TODO: makes this only funcs that return a result
    grExpressionMode = (SupLit("Exprmode") +
                        Group(leVariable)("target") +
                        SupLit("=") +
                        Group(grSubExpression | grExpressionOpcode | leVarOrWord)("subexp"))
    grSubExpression << (SupLit("(") +
                        Group(leVarOrWord | grSubExpression | grExpressionOpcode)("arg1") +
                        oneOf("+ - / * =")("operation") +
                        Group(leVarOrWord | grSubExpression | grExpressionOpcode)("arg2") +
                        SupLit(")"))
    
    # Blocks and larger constructs
    grConstruct = grIncDec | grInlineOperation | grExpressionMode | grJumpGoto | grJumpUnless | grQuirkyFunctions | grFunctions # single-line construct
    grLinePrefix = Optional(grJumpLabel) + Suppress(Optional(grDescummOpcode))
    grLineSuffix = Suppress(Optional(leComment))
    
    # grStartScriptLine is a hack to get around descumm not outputting the extra opcode flags
    grStartScriptLine = Optional(grJumpLabel) + Optional(grDescummOpcode)("descummop") + grFunc_startScript + grLineSuffix
    # grStopObjectCodeLine marks the end of a script
    grStopObjectCodeLine = grLinePrefix + grFunc_stopObjectCode + grLineSuffix
    grNormalLine = grLinePrefix + (grConstruct + Optional(SupLit(";")) | Suppress(leComment)) + grLineSuffix
    grIfLineStart = grLinePrefix + grIfStart + grLineSuffix
    grIfLineElse = grLinePrefix + grIfElse + (Optional(grIfStart)("elseif") ^ SupLit("{")) + grLineSuffix
    grIfLineEnd = grLinePrefix + grIfEnd + grLineSuffix
    
    grWhileLineStart = grLinePrefix + grWhileStart + grLineSuffix
    grWhileLineEnd = grLinePrefix + grWhileEnd + grLineSuffix
    grDoWhileLineStart = grLinePrefix + grDoWhileStart + grLineSuffix
    grDoWhileLineEnd = grLinePrefix + grDoWhileEnd + grLineSuffix
    grForLoopLineStart = grLinePrefix + grForLoopStart + grLineSuffix
    grForLoopLineEnd = grLinePrefix + grForLoopEnd + grLineSuffix
    
    grBlock = Forward()
    # grBlock is optional, because descumm might try to be clever and create "if/else" statements when it's just doing
    #  conditional jumps.
    grIfBlock = grIfLineStart + Optional(OneOrMore(grBlock)) + Optional(OneOrMore(grIfLineElse + ZeroOrMore(grBlock))) + grIfLineEnd
    grWhileBlock = grWhileLineStart + OneOrMore(grBlock) + grWhileLineEnd
    grDoWhileBlock = grDoWhileLineStart + OneOrMore(grBlock) + grDoWhileLineEnd
    grForLoopBlock = grForLoopLineStart + OneOrMore(grBlock) + grForLoopLineEnd
    grBlock << (grNormalLine ^ grStartScriptLine ^ grIfBlock ^ grDoWhileBlock ^ grWhileBlock ^ grForLoopBlock)
    grMainBlock = StringStart() + ZeroOrMore(grPragma) + ZeroOrMore(grBlock) + grStopObjectCodeLine + Optional(SupLit("END")) + StringEnd()
    # This is used for testing, so we don't have to put "stopObjectCode()" at the end of every test
    grMainBlockTesting = StringStart() + ZeroOrMore(grPragma) + ZeroOrMore(grBlock) + Optional(grStopObjectCodeLine) + Optional(SupLit("END")) + StringEnd()

    rootExpression = 'grMainBlock'
    testExpression = 'grMainBlockTesting'


def _unit_test():
    scummgram = GrammarV5()

    for g in scummgram:
        print str(g[0])

if __name__ == '__main__':
    _unit_test()
